/*
Malloc debugger for glibc, by Patrick Reynolds <reynolds@cs.duke.edu>.
Licensed under the GNU Library General Public License; see the file COPYING
for details.
*/

/*#undef GET_CALL_STACK_ACTUALLY_WORKS*/
#define GET_CALL_STACK_ACTUALLY_WORKS
#define MALLOC_DEBUG_DL_SYMBOLS
#define DL_ADDRESS_MASK 0xf0000000

#define _GNU_SOURCE
#define MALLOC_HOOKS
#include <malloc.h>
#include <stdio.h>
#include <features.h>

#ifdef MALLOC_DEBUG_DL_SYMBOLS
#include <dlfcn.h>
#endif /* MALLOC_DEBUG_DL_SYMBOLS */

#if (__GLIBC__ == 2) && (__GLIBC_MINOR__ >= 1)
#define GLIBC2_1
#endif

/* Three function pointers to save the "real" *alloc and free */
#ifdef GLIBC2_1
static __malloc_ptr_t (*saved_malloc) __P ((size_t __size, __const __malloc_ptr_t caller));
static __malloc_ptr_t (*saved_realloc) __P ((__ptr_t __ptr, size_t __size, __const __malloc_ptr_t caller));
static void (*saved_free) __P ((__ptr_t __ptr, __const __malloc_ptr_t caller));
#else
static __ptr_t (*saved_malloc) __P ((size_t __size));
static __ptr_t (*saved_realloc) __P ((__ptr_t __ptr, size_t __size));
static void (*saved_free) __P ((__ptr_t __ptr));
#endif

/* Function pointer to save main() in case my call_stack tracer wants it.
 * Not yet used. */
static int (*saved_main) __P ((int argc, char **argv));

/* Where we write. */
static FILE *log;

extern int main(int argc, char **argv);
static void print_call_stack();

/* hook for malloc() */
#ifdef GLIBC2_1
static __ptr_t ahook(size_t n, __const __malloc_ptr_t caller) {
#else
static __ptr_t ahook(size_t n) {
#endif
  __ptr_t temp;                   /* store return from real malloc */
  __malloc_hook = saved_malloc;   /* restore real malloc */
  temp = malloc(n);               /* call real malloc */
  __malloc_hook = ahook;          /* redirect malloc to me again */
  fprintf(log, "malloc: %d 0x%08lx\n", n, (unsigned long)temp);
  print_call_stack();
  fflush(log);
  return temp;
}

#ifdef GLIBC2_1
static void fhook(__ptr_t p, __const __malloc_ptr_t caller);
#else
static void fhook(__ptr_t p);
#endif

/* hook for realloc(), uses the same strategy as above */
#ifdef GLIBC2_1
static __ptr_t rahook(__ptr_t p, size_t n, __const __malloc_ptr_t caller) {
#else
static __ptr_t rahook(__ptr_t p, size_t n) {
#endif
  __ptr_t temp;
  __realloc_hook = saved_realloc;
  __malloc_hook = saved_malloc;
  __free_hook = saved_free;
  temp = realloc(p, n);
  __realloc_hook = rahook;
  __malloc_hook = ahook;
  __free_hook = fhook;
  fprintf(log, "realloc: %d 0x%08lx 0x%08lx\n",
    n, (unsigned long)temp, (unsigned long)p);
  print_call_stack();
  fflush(log);
  return temp;
}

/* hook for free(), uses the same strategy as above */
#ifdef GLIBC2_1
static void fhook(__ptr_t p, __const __malloc_ptr_t caller) {
#else
static void fhook(__ptr_t p) {
#endif
  __free_hook = saved_free;
  free(p);
  __free_hook = fhook;
  fprintf(log, "free: 0x%08lx\n", (unsigned long)p);
  print_call_stack();
  fflush(log);
}

/* enable malloc tracing.  *ilog is where we want codes dumped. */
void malloc_debug_on(FILE *ilog) {
  if (!ilog) return;
  log = ilog;
  saved_malloc = __malloc_hook;   __malloc_hook = ahook;
  saved_realloc = __realloc_hook; __realloc_hook = rahook;
  saved_free = __free_hook;       __free_hook = fhook;
  saved_main = main;
}

/* disable malloc tracing.  This call is only necessary if you want to
 * halt tracing -before- the program terminates.  Normally, don't bother. */
void malloc_debug_off() {
  __malloc_hook = saved_malloc;
  __realloc_hook = saved_realloc;
  __free_hook = saved_free;
  fclose(log);
}

#ifdef GET_CALL_STACK_ACTUALLY_WORKS
int get_call_stack(void **cs, void *m);

int get_call_stack(void **cs, void *m) {
  int count = 0;
  void *trace;
  __asm__ ("movl %%ebp,%0":"=r"(trace));  /* trace = regs.ebp */
  trace = *(void**)trace;
  trace = *(void**)trace;
  /* continue while 'trace' variable points into code space.  Normally,
     the stack terminates with trace==NULL, but this doesn't always seem
     to be the case.  Comparing against crt0.o or _start would be good,
     except that I don't read in any symbol tables...  This works, if 
     unportably.  */
  while ((int)trace >= 0xb0000000 && (int)trace < 0xc0000000
    && *(void**)(trace + 4) != 0 && *(void**)(trace + 4) != m) {
    cs[count] = *(void**)(trace + 4);
    trace = *(void**)trace;
    count++;
  }
  return count;
}
#endif
  
/* This prints the call_stack (saved %eip's).  I hope to make it print
 * function -names-, too, since that's a lot more useful than addresses. */
void print_call_stack() {
  void *cs[1000];  /* I hope we're not more than 1000 levels in recursion */
  int i, depth = 0;
#ifdef MALLOC_DEBUG_DL_SYMBOLS
  Dl_info dl_info;
#endif /* MALLOC_DEBUG_DL_SYMBOLS */

#ifdef GET_CALL_STACK_ACTUALLY_WORKS
  fprintf(log, "> call stack:\n");
  depth = get_call_stack(cs, saved_main);     /* see get_call_stack.s */
#endif
  for (i=0; i<depth; i++)
#ifdef MALLOC_DEBUG_DL_SYMBOLS
    if (((unsigned long)cs[i] & DL_ADDRESS_MASK) && dladdr(cs[i], &dl_info))
      fprintf(log, ">   0x%08lx (%s)\n", (unsigned long)cs[i], dl_info.dli_sname);
    else
#endif /* MALLOC_DEBUG_DL_SYMBOLS */
      fprintf(log, ">   0x%08lx\n", (unsigned long)cs[i]);
}
